/* On nar:
	$ mpiicc -openmp first-mpi-openmp.c -o first-mpi-openmp
	$ qsub -I -l nodes=4:ppn=8 -q cengc
	$ mpirun -r ssh -np 1 ./first-mpi-openmp # and 2 and 3 ...

   Each MPI process spawns a distinct OpenMP master thread; so limit the
   number of MPI processes to one per node. 
*/ 
#include <omp.h>
#include "mpi.h"
#include <stdio.h>

int main (int argc, char *argv[]) {
	int world_size, rank_id, thread_id, c = 0, thread_count, 
	  total_threads = 0;

/* Set the number of threads to use. Or, you can use the OMP_NUM_THREADS
   environment variable to control how many threads are used.  In the 
   absence of either the number of threads will equal the number of cores
   available on the underlying CPU(s).
*/
//	int _NUM_THREADS = 2;
// 	omp_set_num_threads(_NUM_THREADS);

	MPI_Init(&argc, &argv);
	MPI_Comm_size(MPI_COMM_WORLD, &world_size);
	MPI_Comm_rank(MPI_COMM_WORLD, &rank_id);
	
	if (rank_id == 0) {
		printf("Total number of MPI processes: %d\n", world_size); 
	}
	
	/* This is a parallel OpenMP block executed by each MPI process and 
	   each OpenMP thread.  reduction with + adds all the thread_counts 
	   together.
	*/ 
	#pragma omp parallel reduction(+:c)
	{
		c = 1;
 	}

	/* This is a parallel OpenMP block executed by each MPI process and 
	   each OpenMP thread.  private ensures that thread_id is private to 
	   each thread.
	*/
	#pragma omp parallel private(thread_id) 
	{
		thread_id = omp_get_thread_num();
		printf("rank_id: %d, thread_id: %d\n", rank_id, thread_id); 
	}
	
	/* This is a parallel OpenMP block executed by each MPI process but 
	   only the master OpenMP thread contained within that MPI process.  
	*/ 
	#pragma omp parallel
	{
		#pragma omp master
		{
			thread_count = omp_get_num_threads();
			printf("rank_id: %d, thread_count: %d, c: %d\n", 
			  rank_id, thread_count, c); 
		}
	}
	
	MPI_Reduce(&thread_count, &total_threads, 1, MPI_INT, MPI_SUM, 0,
	  MPI_COMM_WORLD);
	printf("Total number of threads: %d\n", total_threads); 

	MPI_Finalize();
	return(0);
}

